package be.cylab.java.roc;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.nio.charset.StandardCharsets;
import java.util.List;

/**
 * Class to compute statistical information.
 */
public class Statistics {
    private int number_of_elements;

    private int true_positive_counter;
    private int true_negative_counter;
    private int false_positive_counter;
    private int false_negative_counter;

    private double true_positive_rate;
    private double true_negative_rate;
    private double false_positive_rate;
    private double false_negative_rate;

    private double precision;
    private double negative_prediction_value;
    private double false_discovery_rate;
    private double false_omission_rate;
    private double accuracy;
    private double f_1_score;

    /**
     * Constructor.
     *
     * @param points
     * @param threshold
     */
    public Statistics(final List<Point> points,
                      final double threshold) {
        number_of_elements = points.size();

        for (int i = 0; i < points.size(); i++) {
            if (points.get(i).getTrueAlert() == 1
                    && points.get(i).getScore() > threshold) {
                true_positive_counter++;
            } else if (points.get(i).getTrueAlert() == 1
                    && points.get(i).getScore() < threshold) {
                false_negative_counter++;
            } else if (points.get(i).getTrueAlert() == 0
                    && points.get(i).getScore() < threshold) {
                true_negative_counter++;
            } else if (points.get(i).getTrueAlert() == 0
                    && points.get(i).getScore() > threshold) {
                false_positive_counter++;
            }
        }
        computePrecision();
    }

    /**
     * Compute True Positive Rate.
     * Can be called Sensitivity and Recall
     */
    final void computeTruePositiveRate() {
        double positive_condition = true_positive_counter
                + false_negative_counter;
        true_positive_rate = true_positive_counter / positive_condition;
    }

    /**
     * Compute True Negative Rate.
     * Also called Specificity
     */
    final void computeTrueNegativeRate() {
        double negative_condition = true_negative_counter
                + false_positive_counter;
        true_negative_rate = true_negative_counter / negative_condition;
    }

    /**
     * Compute False Negative Rate.
     */
    final void computeFalseNegativeRate() {
        double tmp = false_negative_counter
                + true_positive_counter;
        false_negative_rate = false_negative_counter / tmp;
    }

    /**
     * Compute False Positive Rate.
     * Also called Fall-out
     */
    final void computeFalsePositiveRate() {
        double tmp = false_positive_counter + true_negative_counter;
        false_positive_rate = false_positive_counter / tmp;
    }

    /**
     * Compute Precision.
     * Also called Positive Predictive Value
     */
    final void computePrecision() {
        precision = true_positive_counter / (double)
                (true_positive_counter + false_positive_counter);
    }

    /**
     * Compute Negative Predictive Value.
     */
    final void computeNegativePredictiveValue() {
        negative_prediction_value = true_negative_counter / (double)
                (true_negative_counter + false_negative_counter);
    }

    /**
     * Compute False Discovery Rate.
     */
    final void computeFalseDiscoveryRate() {
        false_discovery_rate = false_positive_counter / (double)
                (false_positive_counter + true_positive_counter);
    }

    /**
     * Compute False Omission Rate.
     */
    final void computeFalseOmissionRate() {
        false_omission_rate = false_negative_counter / (double)
                (false_negative_counter + true_negative_counter);
    }

    /**
     * Compute Accuracy.
     */
    final void computeAccuracy() {
        accuracy = (true_positive_counter + true_negative_counter) / (double)
                number_of_elements;
    }

    /**
     * Compute F1 Score.
     */
    final void computeF1Score() {
        f_1_score = (2 * true_positive_counter) / (double)
                (2 * true_positive_counter + false_positive_counter + false_negative_counter);
    }

    /**
     * Getter number_of_elements.
     *
     * @return
     */
    public final int getNumberOfElements() {
        return number_of_elements;
    }

    /**
     * Getter true_positive_counter.
     *
     * @return
     */
    public final int getTruePositiveCounter() {
        return true_positive_counter;
    }

    /**
     * Getter true_negative_counter.
     *
     * @return
     */
    public final int getTrueNegativeCounter() {
        return true_negative_counter;
    }

    /**
     * Getter false_positive_counter.
     *
     * @return
     */
    public final int getFalsePositiveCounter() {
        return false_positive_counter;
    }

    /**
     * Getter false_negative_counter.
     *
     * @return
     */
    public final int getFalseNegativeCounter() {
        return false_negative_counter;
    }

    /**
     * Getter true_positive_rate.
     *
     * @return
     */
    public final double getTruePositiveRate() {
        computeTruePositiveRate();
        return true_positive_rate;
    }

    /**
     * Getter for Recall (idem True Positive Rate).
     *
     * @return
     */
    public final double getRecall() {
        computeTruePositiveRate();
        return true_positive_rate;
    }

    /**
     * Getter true_negative_rate.
     *
     * @return
     */
    public final double getTrueNegativeRate() {
        computeTrueNegativeRate();
        return true_negative_rate;
    }

    /**
     * Getter false_positive_rate.
     *
     * @return
     */
    public final double getFalsePositiveRate() {
        computeFalsePositiveRate();
        return false_positive_rate;
    }

    /**
     * Getter false_negative_rate.
     *
     * @return
     */
    public final double getFalseNegativeRate() {
        computeFalseNegativeRate();
        return false_negative_rate;
    }

    /**
     * Getter precision.
     * Ne precision computation here because precision can be set in a P-R curve computation.
     *
     * @return
     */
    public final double getPrecision() {
        return precision;
    }

    /**
     * Setter for precision.
     * @param precision
     */
    public final void setPrecision(final double precision) {
        if (precision < 0.0 || precision > 1.0) {
            throw new IllegalArgumentException(
                    "Precision must be between 0 and 1");
        }
        this.precision = precision;
    }

    /**
     * Getter negative_prediction_value.
     *
     * @return
     */
    public final double getNegativePredictionValue() {
        computeNegativePredictiveValue();
        return negative_prediction_value;
    }

    /**
     * Getter false_discovery_rate.
     *
     * @return
     */
    public final double getFalseDiscoveryRate() {
        computeFalseDiscoveryRate();
        return false_discovery_rate;
    }

    /**
     * Getter false_omission_rate.
     *
     * @return
     */
    public final double getFalseOmissionRate() {
        computeFalseOmissionRate();
        return false_omission_rate;
    }

    /**
     * Getter accuracy.
     *
     * @return
     */
    public final double getAccuracy() {
        computeAccuracy();
        return accuracy;
    }

    /**
     * Getter f_1_score.
     *
     * @return
     */
    public final double getF1Score() {
        computeF1Score();
        return f_1_score;
    }

    /**
     * Method to save Stats information in a txt file.
     *
     * @param filename
     * @param threshold
     */
    public final void saveStatsInFile(final String filename,
                                      final double threshold) {
        try (OutputStreamWriter writer = new OutputStreamWriter(
                new FileOutputStream(filename, true),
                StandardCharsets.UTF_8)) {
            writer.write("___________________________________________\n");
            writer.write("|TRIGGER VALUE       :" + threshold + "\n");
            writer.write("|_________________________________________|\n");
            writer.write("WOWA\n\n");

            writer.write("Wowa true positive : "
                    + true_negative_counter + "\n");
            writer.write("Wowa true negative : "
                    + true_negative_counter + "\n");
            writer.write("Wowa false positive : "
                    + false_positive_counter + "\n");
            writer.write("Wowa false negative : "
                    + false_positive_counter + "\n");

            writer.write("Wowa true positive rate : "
                    + true_positive_rate + "\n");
            writer.write("Wowa true negative rate : "
                    + true_negative_rate + "\n");
            writer.write("Wowa false positive rate : "
                    + false_positive_rate + "\n");
            writer.write("Wowa false negative rate : "
                    + false_negative_rate + "\n");

            writer.write("Wowa precision : " + precision + "\n");
            writer.write("Wowa negative predictive value : "
                    + negative_prediction_value + "\n");
            writer.write("Wowa false discovery rate : "
                    + false_discovery_rate + "\n");
            writer.write("Wowa false omission rate : "
                    + false_omission_rate + "\n");
            writer.write("Wowa accuracy : " + accuracy + "\n");
            writer.write("Wowa F1 score : " + f_1_score + "\n");

        } catch (IOException e) {
            e.printStackTrace();
        }
    }
}
